# include <stdlib.h>
# include "gcc-externs.h"
# include "list.h"
# include "sdg.h"
# include "bb.h"
# include "ud.h"
# include "regset.ic"

static int  gotNewReg = 0;
static int  pseudoRegIndex = -1;
List assocList;

void
InitFreePseudoRegList()
{
	gotNewReg = 0;
	pseudoRegIndex = LAST_VIRTUAL_REGISTER+1;
	Init(assocList);
}

rtx
GetAssociatedReg(rtx oldReg)
{
	Iterator     iter;
	Association *tmp;

	tmp = (Association *)InitIter(iter, assocList);
	while (tmp != NULL) {
		if (tmp->oldReg == oldReg)
			return tmp->newReg;
		tmp = (Association *)GetNext(iter);
	}

	return (rtx)NULL;
}

rtx
GetReplacementReg(rtx oldReg)
{
/* Try to get a free pseudo register between 'FIRST_PSEUDO_REGISTER' 
 * and 'max_regno' that is unused.  If available, create a rtx for
 * this and return. */

	extern regset regs_live_at_setjmp;

	enum machine_mode mode = GET_MODE(oldReg);
	rtx  		    newReg;
	Association *newAssoc;
	
			/* If a pseudo was allocated as a replacement
			 * for oldReg previously, return it */
	newReg = GetAssociatedReg(oldReg);
	if (newReg != NULL)
		return newReg;

			/* For pseudos which are live across 'setjmp's,
			 * 'reg_basic_block[pseudoRegIndex]' is set 
			 * to REG_BLOCK_UNKNOWN */
	for ( ; pseudoRegIndex < max_regno; pseudoRegIndex++)
		if (   (GET_CODE(regno_reg_rtx[pseudoRegIndex]) == REG)
			&& (reg_basic_block[pseudoRegIndex] == REG_BLOCK_UNKNOWN)
			&& !BitIsSet(regs_live_at_setjmp, pseudoRegIndex))
			break;

	if (pseudoRegIndex < max_regno) {
			/* Found a scavenged replacement pseudo */
		newReg = regno_reg_rtx[pseudoRegIndex];
		PUT_MODE(newReg, mode);
	}
	else {
			/* Have run out of scavenged replacement pseudos */
		gotNewReg = 1;
		newReg = gen_reg_rtx(mode);
		max_regno++;
	}

	pseudoRegIndex++;
	newAssoc = (Association *)malloc(sizeof(Association));
	newAssoc->oldReg = oldReg;
	newAssoc->newReg = newReg;
	Append(&assocList, newAssoc);

	return newReg;
}

void
UpdateDataFlowInformation(int oldRegNo, int newRegNo)
{
/* Copy all the dataflow information that was valid 
 * for 'oldReg' for 'newReg'.  Though this will be
 * conservative, this will not lead to incorrect
 * results.  By doing some analysis, better estimates
 * can be obtained, but since this case occurs infrequently,
 * not worth the effort */

	int i;

	reg_basic_block[newRegNo] = reg_basic_block[oldRegNo];
	reg_n_refs[newRegNo] = reg_n_refs[oldRegNo];
	reg_n_sets[newRegNo] = reg_n_sets[oldRegNo];
	reg_n_deaths[newRegNo] = reg_n_deaths[oldRegNo];
	reg_n_calls_crossed[newRegNo] = reg_n_calls_crossed[oldRegNo];
	reg_live_length[newRegNo] = reg_live_length[oldRegNo];
	reg_changes_size[newRegNo] = reg_changes_size[oldRegNo];
	regno_first_uid[newRegNo] = regno_first_uid[oldRegNo];
	regno_last_uid[newRegNo] = regno_last_uid[oldRegNo];
	regno_pointer_flag[newRegNo] = regno_pointer_flag[oldRegNo];
	regno_last_note_uid[newRegNo] = 0;
	for (i = 0; i < n_basic_blocks; i++)
		if (BitIsSet(basic_block_live_at_start[i], oldRegNo))
			SetBit(basic_block_live_at_start[i], newRegNo);
}

void
ChangeRegInLdStSubsystem(rtx oldReg, rtx newReg)
{
/* Replace all occurences of 'oldReg' in instructions
 * of the ld-st subsystem by 'newReg' */

/* Am not using the 'dest' and 'src' pointers because
 * slicing-with-copy-insertion would have inserted copies
 * in the graph without modifying the src and dst pointers */

	SdgNode *tmp = sdg.head;
	rtx      insn, pat;

	extern void ReplaceReg(rtx expr, rtx *exprPtr, rtx oldReg, rtx newReg);
	extern void ReplaceRegInInstr(rtx insn, rtx oldReg, rtx newReg);

		/* Firstly, update data flow information */
	UpdateDataFlowInformation(REGNO(oldReg), REGNO(newReg));

		/* Now, for all instructions in the ldst subsystem
		 * which use 'oldReg', replace it by 'newReg' */

	while (tmp != NULL) {
		if (tmp->inLdSt) {
			insn = tmp->instr;
			if (!tmp->isMemOp)
				ReplaceRegInInstr(insn, oldReg, newReg);
			else if (tmp->isLoadOp) {
					/* Any use of the old register in the address
					 * has to be replaced by the new register */
				pat = (GET_CODE(insn) == SET) ? insn : PATTERN(insn);
				ReplaceReg(SET_SRC(pat), &SET_SRC(pat), oldReg, newReg);

					/* The destination register should be replaced only if
					 * the register is in the ldst subsystem.  This can happen
					 * only when 'tmp' is the "load-value" node and is in LdSt
					 * 'tmp' is the load-value node if !tmp->isAddr */
				if (!tmp->isAddr)
					ReplaceReg(SET_DEST(pat), &SET_DEST(pat), oldReg, newReg);
			}
			else if (tmp->isStoreOp) {
					/* Any use of the old register in the address
					 * has to be replaced by the new register. */
				pat = (GET_CODE(insn) == SET) ? insn : PATTERN(insn);
				ReplaceReg(SET_DEST(pat), &SET_DEST(pat), oldReg, newReg);

					/* The source register should be replaced only if
					 * the register is in the ldst subsystem.  This can
					 * happen only when 'tmp' is the "store-value" node and
					 * is in LdSt */
				if (tmp->isStoreVal)
					ReplaceReg(SET_SRC(pat), &SET_SRC(pat), oldReg, newReg);
			}
		}

		tmp = tmp->next;
	}
}

# ifdef __GNUC__
inline
# endif
rtx
DEST_REG(rtx instr)
{
	rtx reg = (GET_CODE(instr) == SET ? SET_DEST(instr) 
									  : (SET_DEST(PATTERN(instr))));
	while (GET_CODE(reg) == SUBREG)
		reg = SUBREG_REG(reg);

	return reg;
}

void
CheckForRegsInBothSubsystems()
{
/* The way 'gcc' generates code, in some cases, the same pseudo reg
 * might be used for disjoint uses of a value.  Since slicing is done
 * based on values and not names of pseudo-regs, a particular use of a
 * pseudo reg might be in the comp. subsystem and another might be in
 * the ld-st subsystem.  Identify such pseudo regs and introduce a new
 * pseudo reg for all uses in the ld-st subsystem */ 

	int   i;
	Defn *defn;
	rtx   oldReg, newReg;
	int   numChanges = 0;

	extern void life_analysis(rtx first_insn, int nregs, int fix_info);
	extern regset regsInComp;

	FreeList(&assocList);
	Init(assocList);

			/* For each defn, check whether this defn supplies
			 * any instr in the Ldst subsystem.  If the dest.
			 * of the defn. is in the Comp subsystem, then
			 * this reg. is being used in both subsystems */
	for (i = 0; i < numDefns; i++) {
		defn = GetDefn(i);
		defn->regNo = REGNO(DEST_REG(defn->insn));
		if ((defn->regNo != 0) && !IsEmpty(defn->node->dsts)) {
			if (defn->node->inLdSt && BitIsSet(regsInComp, defn->regNo)) {
				fprintf(stderr, 
						"Pseudo-register %d in both subsystems", 
						defn->regNo);
					/* Try to get a free pseudo register between
					 * 'FIRST_PSEUDO_REGISTER' and 'max_regno' that
					 * is unused.  If available, replace 'defn->regNo'
					 * by the new register in all instructions of
					 * the ldst subsystem. */
				oldReg = SET_DEST(defn->insn);
				newReg = GetReplacementReg(oldReg);
				if (newReg != NULL) {
					fprintf(stderr, " -- Replacing %d by %d\n", 
							REGNO(oldReg), REGNO(newReg));
					ChangeRegInLdStSubsystem(oldReg, newReg);
					numChanges++;
				}
				else
					fprintf(stderr, "\n");
			}
		}
	}

		/* Redo flow analysis to take into account the
		 * introduction of the new registers. */
	flow_analysis(get_insns(), max_regno, flow_dump_file);
	FreeList(&assocList);
}
