# include <stdlib.h>
# include <assert.h>
# include "gcc-externs.h"
# include "list.h"
# include "sdg.h"
# include "bb.h"
# include "ud.h"
# include "regset.ic"

		/* The list of instructions from which slicing begins */
List sliceNodeList;

void
ClearAllVisitedBits()
{
	SdgNode *node = sdg.head;

	while (node != NULL) {
		node->visited = 0;
		node = node->next;
	}
}

void
FindReachableNodes(SdgNode *node)
{
/* Continue slicing the graph with respect to 'node'
 * both forwards and backwards.  Since we start off
 * this process with load/store nodes, the resulting
 * slice will be all those instructions which will
 * execute in the ld-st subsystem */

	Iterator  iter;
	SdgNode  *tmp;

	assert(!node->visited);

	node->inLdSt = 1;
	node->visited = 1;

		/* For each of the sources, slice with respect to it.
		 * Finds the backward slice. */
	for (tmp = (SdgNode *)InitIter(iter, node->srcs);
	     tmp != NULL;
		 tmp = (SdgNode *)GetNext(iter))
	{
		if (!tmp->visited)
			FindReachableNodes(tmp);
	}

	for (tmp = (SdgNode *)InitIter(iter, node->dsts);
	     tmp != NULL;
		 tmp = (SdgNode *)GetNext(iter))
	{
		if (!tmp->visited)
			FindReachableNodes(tmp);
	}
}

void
SliceDependencyGraph()
{
/* 'sliceNodeList' is the list of all load/store instructions
 * AND all nodes which are already determined to be in the load/store slice
 * (instructions which pass parameters to functions, for example) */

	SdgNode *sliceNode;

# ifdef PLAIN_SLICING
	SdgNode *node;

	for (node = sdg.head; node != NULL; node = node->next) {
		if (node->dstUsesIntArgReg || node->dstUsesIntRetReg) {
			node->inLdSt = 1;
			node->dstUsesHardIntReg = 1;
			if (!node->isMemOp)
				Append(&sliceNodeList, node);
		}
		else if (node->srcUsesIntArgReg || node->srcUsesIntRetReg) {
			node->inLdSt = 1;
			node->srcUsesHardIntReg = 1;
			if (!node->isMemOp)
				Append(&sliceNodeList, node);
		}
	}
# endif

	for (sliceNode = RemoveHead(&sliceNodeList);
	     sliceNode != NULL;
		 sliceNode = RemoveHead(&sliceNodeList))
	{
		if (!sliceNode->visited)
			FindReachableNodes(sliceNode);
	}
}
