# ifndef __SDG_H_
# define __SDG_H_

# include "types.h"

struct SdgNode {
	int      id;			  /* Id of this node */
	rtx      instr;			  /* The RTL instruction corr. to this node. */
	int      inCompoundRtx:1; /* Is this insn inside a parallel/sequence rtx? */
	rtx      compoundRtx;	  /* If 'inCompoundRtx', the compound rtx */
	AdjList  dsts;			  /* The insns which use the value produced by me */
	AdjList  srcs;			  /* The insns which produce one of my operands */
	SdgNode *next;			  /* Next node in the graph */
	float    nodeWeight;	  /* If this node is in basic block 'b', then
	                           * nodeWeight = prob(b) * 5^num_nested_loops(b) */
	int      inLdSt:1;  	  /* Is this node in the 'ld-st' subsystem */
	int      isMoveOp:1;	  /* Is this a reg-reg move? */
	int      isACall:1;		  /* Is this a call instruction? */
	int      isMemOp:1;		  /* Is this a load/store operation? */
	int      isLoadOp:1;	  /* Is this a load operation? */
	int      isAddr:1;		  /* Does this node represent a load/store addr? */
	int      isStoreOp:1;	  /* Is this a store operation? */
	int      isStoreVal:1;	  /* Does this node represent a store value? */
	int      isAConvNode:1;   /* Is this node a conversion node?
	                           * Yes => values are communicated between the two
							   * register files since they convert values
							   * between floating point and integer formats.
							   * These nodes have a dual personality.
							   * To sources, they are stores.
							   * To destinations, they are loads. */
	int      srcUsesHardIntReg:1;/* Does the source use an int hard reg? */
	int      dstUsesHardIntReg:1;/* Does the destination use an int hard reg? */
	                          /* For non-copy/non-memory-op nodes, it doesn't 
							   * matter whether the destination/source uses
							   * an int hard reg.  If either one uses it, then
							   * the node and along with it all of its source
							   * and destination nodes go into the ld-st slice.
							   * However, for a load, if the destination is
							   * a hard int reg, then all destinations go into
							   * the ld-st slice.  For a store, if the source
							   * is a hard int reg, then all sources go into
							   * the ld-st slice.  A copy node as mentioned
							   * below is both a load-store */
	int      srcUsesIntArgReg:1;/* Does the source use an int argument reg? */
	int      dstUsesIntArgReg:1;/* Does the dest. use an int arg reg? */
	int      srcUsesIntRetReg:1;/* Does the source use an int return reg? */
	int      dstUsesIntRetReg:1;/* Does the dest. use an int return reg? */
	int      visited:1;       /* Has this node been visisted during the
	                           * current analysis phase */
# ifdef COPYOPT_SLICING
	int      inProfile:1;     /* Was this node in the BB profile? */
	int      insertedInList:1;/* A marker to avoid duplicates */
	int      copiesAreNeeded:1;
							  /* If this is a load/store, would copies be
							   * needed if children/parents are going to be
							   * in comp? */
	int      mightBeInLdst:1; /* Is this node temporarily assigned to LdSt */
	int      copiesMightBeNeeded:1; /* If 'mightBeInLdst', are copies needed? */
	int      copyInserted:1;  /* If 'copiesAreNeeded', have the copies been 
							   * inserted? */
	int      shouldDuplicate:1;/* Instead of creating a copy, is is better to
	                           * to duplicate this instr? */
	float    duplicatingCost; /* Cost of duplicating this instr. */
	float    copyingCost;	  /* Cost of a copy of this instr's dest reg */
	rtx      replacementReg;  /* If 'copiesAreNeeded', the replacement reg 
							   * for this instruction's destination */
# endif
};

	/* A graph is just a list of nodes */
typedef struct Graph {
	SdgNode *head;
	SdgNode *tail;
} Graph;

extern regset  *regsDefinedInBB;
extern SdgNode *currSdgNode;
extern SdgNode *dummyArgNodes[];
extern Graph    sdg;

extern SdgNode *AllocNewSDGnode(rtx insn);
extern SdgNode *InsertInsnIntoSDG(rtx insn);

# endif
